import struct
from typing import List

from .constants import *


class Header:
    """Read and provide interface to header"""

    def __init__(self, file):
        byte_buffer = file.read(5)  # "HFST\0"
        if struct.unpack_from("<5s", byte_buffer, 0)[0] == b'HFST\x00':
            # just ignore any hfst3 header
            remaining = struct.unpack_from("<H", file.read(3), 0)[0]
            self._handle_hfst3_header(file, remaining)
            byte_buffer = file.read(56)  # 2 unsigned shorts, 4 unsigned ints and 9 uint-bools
        else:
            byte_buffer = byte_buffer + file.read(56 - 5)
        self.number_of_input_symbols = struct.unpack_from("<H", byte_buffer, 0)[0]
        self.number_of_symbols = struct.unpack_from("<H", byte_buffer, 2)[0]
        self.size_of_transition_index_table = struct.unpack_from("<I", byte_buffer, 4)[0]
        self.size_of_transition_target_table = struct.unpack_from("<I", byte_buffer, 8)[0]
        self.number_of_states = struct.unpack_from("<I", byte_buffer, 12)[0]
        self.number_of_transitions = struct.unpack_from("<I", byte_buffer, 16)[0]
        self.weighted = struct.unpack_from("<I", byte_buffer, 20)[0] != 0
        self.deterministic = struct.unpack_from("<I", byte_buffer, 24)[0] != 0
        self.input_deterministic = struct.unpack_from("<I", byte_buffer, 28)[0] != 0
        self.minimized = struct.unpack_from("<I", byte_buffer, 32)[0] != 0
        self.cyclic = struct.unpack_from("<I", byte_buffer, 36)[0] != 0
        self.has_epsilon_epsilon_transitions = struct.unpack_from("<I", byte_buffer, 40)[0] != 0
        self.has_input_epsilon_transitions = struct.unpack_from("<I", byte_buffer, 44)[0] != 0
        self.has_input_epsilon_cycles = struct.unpack_from("<I", byte_buffer, 48)[0] != 0
        self.has_unweighted_input_epsilon_cycles = struct.unpack_from("<I", byte_buffer, 52)[0] != 0

    def _handle_hfst3_header(self, file, remaining):
        chars = struct.unpack_from("<" + str(remaining) + "c",
                                   file.read(remaining), 0)
        # assume the h3-header doesn't say anything surprising for now


class Alphabet:
    """Read and provide interface to alphabet"""

    def __init__(self, file, number_of_symbols):
        key_table = []  # type: List[str]
        flag_diacritic_operations = dict()  # of symbol numbers to string triples

        for x in range(number_of_symbols):

            symbol = b""

            while True:
                byte = file.read(1)
                if byte == b'\0':  # a symbol has ended
                    # symbol = unicode(symbol, "utf-8")
                    symbol_str = symbol.decode(encoding='utf-8')
                    if len(symbol_str) > 4 and symbol_str[0] == '@' and \
                            symbol_str[-1] == '@' and symbol_str[2] == '.' and symbol_str[1] in "PNRDCU":
                        # this is a flag diacritic
                        op = feat = val = ""
                        parts = symbol_str[1:-1].split('.')
                        if len(parts) == 2:
                            op, feat = parts
                        elif len(parts) == 3:
                            op, feat, val = parts
                        else:

                            key_table.append(symbol_str)
                            break
                        flag_diacritic_operations[x] = FlagDiacriticOperation(op, feat, val)
                        key_table.append("")
                        break
                    key_table.append(symbol_str)
                    break
                symbol += byte
        key_table[0] = ""
        self.key_table = key_table
        self.flagDiacriticOperations = flag_diacritic_operations


class LetterTrie:
    """Insert and prefix-retrieve string / symbol number pairs"""

    class Node:

        def __init__(self):
            self.symbols = dict()
            self.children = dict()

        def add(self, string, symbolNumber):
            """
            Add string to trie, having it resolve to symbolNumber
            """
            if len(string) > 1:
                if not (string[0] in self.children):
                    self.children[string[0]] = self.__class__()  # instantiate a new node
                self.children[string[0]].add(string[1:], symbolNumber)
            elif len(string) == 1:
                self.symbols[string[0]] = symbolNumber
            else:
                self.symbols[string] = symbolNumber

        def find(self, indexstring):
            """
            Find symbol number corresponding to longest match in indexstring
            (starting from the position held by indexstring.pos)
            """
            if indexstring.pos >= len(indexstring.s):
                return NO_SYMBOL_NUMBER
            current = indexstring.get()
            indexstring.pos += 1
            if not (current in self.children):
                if not (current in self.symbols):
                    indexstring.pos -= 1
                    return NO_SYMBOL_NUMBER
                return self.symbols[current]
            temp = self.children[current].find(indexstring)
            if temp == NO_SYMBOL_NUMBER:
                if not (current in self.symbols):
                    indexstring.pos -= 1
                    return NO_SYMBOL_NUMBER
                return self.symbols[current]
            return temp

    def __init__(self):
        self.root = self.Node()

    def addString(self, string, symbolNumber):
        self.root.add(string, symbolNumber)

    def findKey(self, indexstring):
        return self.root.find(indexstring)


class Indexlist:
    """Utility class to keep track of where we are in a list"""

    def __init__(self, items=[]):
        self.s = list(items)
        self.pos = 0

    def get(self, adjustment=0):
        return self.s[self.pos + adjustment]

    def put(self, val, adjustment=0):
        if (self.pos + adjustment) < len(self.s):
            self.s[self.pos + adjustment] = val
        else:
            self.s.append(val)

    def save(self):
        self.temp = self.pos

    def restore(self):
        self.pos = self.temp

    def last(self):
        if len(self.s) > 0:
            return self.s[-1]
        return None


class FlagDiacriticOperation:
    """Represents one flag diacritic operation"""

    def __init__(self, op, feat, val):
        self.operation = op
        self.feature = feat
        self.value = val


class FlagDiacriticStateStack:
    """The combined state for all flag diacritics"""

    def __init__(self):
        self.stack = [dict()]

    def clear(self):
        self.stack = [dict()]

    def pop(self):
        self.stack.pop()

    def duplicate(self):
        self.stack.append(self.stack[-1].copy())

    def push(self, flag_diacritic):
        """
        Attempt to modify flag diacritic state stack. If successful, push new
        state and return True. Otherwise return False.
        """
        if flag_diacritic.operation == 'P':  # positive set
            self.duplicate()
            self.stack[-1][flag_diacritic.feature] = (flag_diacritic.value, True)
            return True
        if flag_diacritic.operation == 'N':  # negative set
            self.duplicate()
            self.stack[-1][flag_diacritic.feature] = (flag_diacritic.value, False)
            return True
        if flag_diacritic.operation == 'R':  # require
            if flag_diacritic.value == '':  # empty require
                if self.stack[-1].get(flag_diacritic.feature) is None:
                    return False  # empty require = require nonempty value
                else:
                    self.duplicate()
                    return True
            if self.stack[-1].get(flag_diacritic.feature) == (flag_diacritic.value, True):
                self.duplicate()
                return True
            return False
        if flag_diacritic.operation == 'D':  # disallow
            if flag_diacritic.value == '':  # empty disallow
                if self.stack[-1].get(flag_diacritic.feature) is None:
                    self.duplicate()
                    return True
                else:
                    return False
            if self.stack[-1].get(flag_diacritic.feature) == (flag_diacritic.value, True):
                return False
            self.duplicate()
            return True
        if flag_diacritic.operation == 'C':  # clear
            self.duplicate()
            if flag_diacritic.feature in self.stack[-1]:
                del self.stack[-1][flag_diacritic.feature]
            return True
        if flag_diacritic.operation == 'U':  # unification
            if not flag_diacritic.feature in self.stack[-1] or \
                    self.stack[-1][flag_diacritic.feature] == (flag_diacritic.value, True) or \
                    (self.stack[-1][flag_diacritic.feature][1] == False and \
                     self.stack[-1][flag_diacritic.feature][0] != flag_diacritic.value):
                self.duplicate()
                self.stack[-1][flag_diacritic.feature] = (flag_diacritic.value, True)
                return True
            return False


def match(transition_symbol, input_symbol):
    """Utility function to check whether we want to traverse a transition/index"""
    if transition_symbol == NO_SYMBOL_NUMBER:
        return False
    if input_symbol == NO_SYMBOL_NUMBER:
        return True
    return transition_symbol == input_symbol



#!/usr/bin/env python3
import argparse
import atexit
import queue
import shutil
import subprocess
import sys
import threading
import time
from pathlib import Path
from typing import List, Iterable, Dict, Set, Tuple, Union
from weakref import WeakSet

from hfstol.shared import Header, Alphabet
from hfstol.transducer import Transducer

PathLike = Union[Path, str]  # python 3.5 compatibility


class HFSTOL:
    # All HFSTOL objects that gets created are added to this
    # weakset. If Python begins tearing down and an HFSTOL process is still in
    # referenced in memory, then all its processes are automatically
    # terminated, before the garbage collector gets involved.
    # _terminate_all_active_hfstol_processes().
    # Once an HFSTOL instance is no longer referenced any more, it is
    # automagically~~ removed from this set!
    ACTIVE_HFSTOL_INSTANCES = WeakSet()  # type: WeakSet

    def __init__(self, header, alphabet, transducer, hfstol_file_path):
        """
        Read a transducer from filename
        """
        self._header = header
        self._alphabet = alphabet
        self._transducer = transducer

        self._hfstol_exe_path = shutil.which("hfst-optimized-lookup")
        self._hfstol_file_path = hfstol_file_path

        self._hfstol_processes = []  # type: List[subprocess.Popen]

        # Keep track of this object, so that if it hasn't been garbage
        # collected by the time the Python process is shutting down, we
        # can terminate all processes before Python starts deleting all the
        # references we need to do so!
        self.ACTIVE_HFSTOL_INSTANCES.add(self)

    def feed(
        self, surface_form: str, concat: bool = True
    ) -> Tuple[Tuple[str, ...], ...]:
        """
        feed surface form to hfst

        :param surface_form: the surface form
        :param concat: whether to concatenate single characters

            example output for `surface_form` = 'niskak', with `crk-descriptive-analyzer.hfstol`
            - True: (('niska', '+N', '+A', '+Pl'), ('nîskâw', '+V', '+II', '+II', '+Cnj', '+Prs', '+3Sg'))
            - False: (('n', 'i', 's', 'k', 'a', '+N', '+A', '+Pl'), ('n', 'î', 's', 'k', 'â', 'w', '+V', '+II', '+II', '+Cnj', '+Prs', '+3Sg'))

            example output for `surface_form` = 'niska+N+A+Pl' with `crk-normative-generator.hfstol`
            - True: (('niskak',),)
            - False: (('n', 'i', 's', 'k', 'a', 'k'),)

            example output for `surface_form` = 'niska+N+A+Pl' with `crk-normative-generator.hfstol` (an inflection that has two spellings)
            - True: (('kinipânaw',), ('kinipânânaw',))
            -False: (('k', 'i', 'n', 'i', 'p', 'â', 'n', 'a', 'w'), ('k', 'i', 'n', 'i', 'p', 'â', 'n', 'â', 'n', 'a', 'w'))
        """
        if surface_form == "":
            return tuple()
        if self._transducer.analyze(surface_form):
            if concat:
                return tuple(concat_res(i) for i in self._transducer.output_list)
            else:
                return tuple(tuple(i) for i in self._transducer.output_list)
        else:
            return tuple()

    def feed_in_bulk(
        self, surface_forms: Iterable[str], concat=True
    ) -> Dict[str, Set[Tuple[str, ...]]]:
        """
        feed a multiple of surface forms to hfst at once

        :param surface_forms:
        :return: a dictionary with keys being each surface form fed in, values being their corresponding deep forms
        """
        res = {
            string: set() for string in surface_forms
        }  # type: Dict[str, Set[Tuple[str,...]]]
        for string in surface_forms:
            res[string] |= set(tuple(self.feed(string, concat)))
        return res

    def feed_in_bulk_fast(
        self, strings: Iterable[str], multi_process: int = 1
    ) -> Dict[str, Set[str]]:
        """
        calls `hfstol-optimized-lookup`. Evaluation is magnitudes faster. Note the generated symbols will all be all concatenated.
        e.g. instead of ['n', 'i', 's', 'k', 'a', '+N', '+A', '+Pl'] it returns ['niska+N+A+Pl']

        :keyword multi_process: Defaults to 1. Specify how many parallel processes you want to speed up computation. A rule is to have processes at most your machine core count.
        """

        if self._hfstol_exe_path is None:
            raise ImportError(
                "hfst-optimized-lookup is not installed.\n"
                "Please install the HFST suite on your system "
                "before using hfstol.\n"
                "See: https://github.com/hfst/hfst#installation"
            )

        while len(self._hfstol_processes) < multi_process:
            proc = subprocess.Popen(
                args=[
                    str(self._hfstol_exe_path),
                    "--quiet",
                    "--pipe-mode",
                    str(self._hfstol_file_path),
                ],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=False,
                universal_newlines=True,
                bufsize=-1,
            )
            self._hfstol_processes.append(proc)

        return self._call_hfstol(list(strings), multi_process)

    @classmethod
    def from_file(cls, filename: PathLike):
        """
        :param filename: the `.hfstol` file
        :return: an `HFSTOL` instance, which you can use to convert surface forms to deep forms
        """
        handle = open(str(filename), "rb")
        header = Header(handle)
        alphabet = Alphabet(handle, header.number_of_symbols)
        if header.weighted:
            raise NotImplementedError("weighted hfstol is not implemented")
            # todo: refer to https://github.com/hfst/hfst-optimized-lookup/tree/master/hfst-optimized-lookup-python/hfst_lookup and implement
            # self.transducer = TransducerW(handle, self.header, self.alphabet)
        else:
            transducer = Transducer(handle, header, alphabet)
        handle.close()
        return cls(header, alphabet, transducer, filename)

    def _call_hfstol(
        self, inputs: List[str], multi_process: int
    ) -> Dict[str, Set[str]]:
        """
        call hfstol and return concatenated results
        """

        length = len(inputs)
        words_per_process = [
            inputs[length * i // multi_process : length * (i + 1) // multi_process]
            for i in range(multi_process)
        ]

        message_queue = queue.Queue()  # type: queue.Queue

        def interact_with_process(
            p: subprocess.Popen, mq: queue.Queue, words: List[str]
        ):
            def _write_lines(stdin, lines):
                for line in lines:
                    stdin.write(line + "\n")
                p.stdin.flush()

            received_count = 0
            old_line = ""
            threading.Thread(target=_write_lines, args=(p.stdin, words)).start()

            while received_count < len(words):
                chars = p.stdout.readline()
                if chars != '':
                    line = chars.rstrip()
                else:
                    time.sleep(0.01)
                    continue

                if line != '':
                    mq.put(line)
                elif old_line:
                    received_count += 1
                old_line = line
            mq.put(None)

        for i, proc in enumerate(self._hfstol_processes[:multi_process]):
            threading.Thread(
                target=interact_with_process,
                args=(proc, message_queue, words_per_process[i]),
            ).start()

        results = {original: set() for original in inputs}  # type: Dict[str, Set[str]]

        ended_count = 0

        while ended_count != multi_process:
            line = message_queue.get()
            if line is None:
                ended_count += 1
                continue
            # Each line will be in this form:
            #    <original_input>\t<res>
            # e.g. in the case of analyzer file
            # nôhkominân    nôhkom+N+A+D+Px1Pl+Sg
            # e.g. in the case of generator file
            # nôhkom+N+A+D+Px1Pl+Sg     nôhkominân
            # If the hfstol can't compute, the transduction will have +?:
            # e.g.,
            #   sadijfijfe	sadijfijfe	+?
            original_input, res, *rest = line.split("\t")

            if "+?" not in rest and "+?" not in res:
                results[original_input].add(res)

        return results

    def __del__(self):
        self._terminate_all_processes()

    def _terminate_all_processes(self):
        """
        Terminates all active hfst-optimized-lookup processes created by this
        instance.
        """
        while self._hfstol_processes:
            proc = self._hfstol_processes.pop()
            proc.terminate()

    @staticmethod
    @atexit.register
    def _terminate_all_active_hfstol_processes():
        """
        Terminates any hfst-optimized-lookup process that have not yet been
        garbage collected.
        """
        for instance in HFSTOL.ACTIVE_HFSTOL_INSTANCES:
            instance._terminate_all_processes()


def cmd():
    parser = argparse.ArgumentParser(description="hfst-optimized-lookup in python")

    parser.add_argument(
        "filename", action="store", type=str, help="The file path of the .hfstol file"
    )

    argv = parser.parse_args()

    transducerfile = open(argv.filename, "rb")
    header = Header(transducerfile)
    print("header read")
    alphabet = Alphabet(transducerfile, header.number_of_symbols)
    print("alphabet read")
    if header.weighted:
        # transducer = TransducerW(transducerfile, header, alphabet)
        # todo: refer to https://github.com/hfst/hfst-optimized-lookup/tree/master/hfst-optimized-lookup-python/hfst_lookup and implement
        raise NotImplementedError("weighted hfstol is not implemented")
    else:
        transducer = Transducer(transducerfile, header, alphabet)
    print("transducer ready")
    print()

    while True:
        try:
            string = input()
        except EOFError:
            sys.exit(0)
        print(string + ":")
        if transducer.analyze(string):
            transducer.printAnalyses()
            print()
        else:
            # tokenization failed
            pass


def concat_res(i) -> Tuple[str, ...]:
    res = []  # type: List[str]
    buffer = ""

    encountered = False
    for symbol in i:
        # print(symbol)
        if len(symbol) == 1:
            encountered = True
            buffer += symbol
        else:
            if encountered:
                res.append(buffer)
                encountered = False
            res.append(symbol)
    if encountered:
        res.append(buffer)
    return tuple(res)


if __name__ == "__main__":
    # just to provide a way to test the cmd entry point
    cmd()
